#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>

#define GST_ERR 1

#define GST_NARGS 3

#define SHIFT 11

void circl_shl(int *val);
void substitution(int s_tab[4][256], int *val);
void create_stab(char src[8][16], int [4][256]);
void crypto_transformation();

															/*test GOST key. */
char test_s_blk[8][16] = { \
{4, 10, 9, 2, 13, 8, 0, 14, 6, 11, 1, 12, 7, 15, 5, 3}, {14, 11, 4, 12, 6, 13, 15, 10, 2, 3, 8, 1, 0, 7, 5, 9}, \
{5, 8, 1, 13, 10, 3, 4, 2, 14, 15, 12, 7, 6, 0, 9, 11}, {7, 13, 10, 1, 0, 8, 9, 15, 14, 4, 6, 12, 11, 2, 5, 3}, \
{6, 12, 7, 1, 5, 15, 13, 8, 4, 10, 9, 14, 0, 3, 11, 2}, {4, 11, 10, 0, 7, 2, 1, 13, 3, 6, 8, 5, 9, 12, 15, 14}, \
{13, 11, 4, 1, 3, 15, 5, 9, 0, 10, 14, 7, 6, 8, 2, 12,}, {1,15, 13, 0, 5, 7, 10, 4, 9, 2, 3, 14, 6, 11, 8, 12}};

int GOST_key[8];											/* Massive for 256 bit gost key. */
int word[2];												/* Massive for A and B(that sit on pipe) */

int stab[4][256];

int main(int argc, char **argv)
{
	int sourcefilesize;
	int fd_dest, fd_src, fd_key, res;
	struct stat st;
	
	if (argc != GST_NARGS + 1) {
		printf("\nusage: program source destination key\n ");
		exit(GST_ERR);					// функция вывода справки.
	}

	fd_src = open(argv[1],O_RDONLY);								/* Opening source file. */
	if (fd_src == -1) {
		perror("OPEN_SRC:");
		exit(GST_ERR);
	}
	res = fstat(fd_src, &st);
	if (res != 0) {
		perror("FSTAT:");
		exit(GST_ERR);
	}
	if (!S_ISREG(st.st_mode)) {									/* Check type of file. */
		fprintf(stderr,"Its not a regular file\n");
		close(fd_src);
		exit(1);
	}

	sourcefilesize = st.st_size;									/* Memorise count of bytes in sourcefile. */

	fd_dest = open(argv[2],O_RDWR | O_CREAT | O_TRUNC,S_IRWXU );						/* Opening destination file. */
	if (fd_dest == -1) {
		perror("OPEN_DEST:");
		exit(GST_ERR);
	}
	res = fstat(fd_dest, &st);	
	if (res != 0) {
		perror("FSTAT:");
		exit(GST_ERR);
	}
	if (!S_ISREG(st.st_mode)) {									/* Check type of file. */
		fprintf(stderr,"Its not a regular file\n");
		close(fd_src);
		exit(1);
	}
	
	fd_key = open(argv[3], O_RDONLY);								/* Opening key file. */
	if (fd_key == -1) {
		perror("OPEN_KEY:");
		exit(GST_ERR);
	}
	res = fstat(fd_key, &st);	
	if (res != 0) {
		perror("FSTAT:");
		exit(GST_ERR);
	}
	if (!S_ISREG(st.st_mode)) {	//проверка на тип
		fprintf(stderr,"Its not a regular file\n");
		close(fd_src);
		exit(1);
	}
	if ( st.st_size < 32) {
		fprintf(stderr, "key is too small ;( try again;)\n");
		exit(1);
	}
	
	
	res = read(fd_key, GOST_key, 32);								/* Trying to read key. */
	if ( res == -1){
		perror("FSTAT:");
		exit(GST_ERR);
	} else if (res < 32) {
		fprintf(stderr, "key READ error. Try again;)\n");
		exit(1);
	}
	
	create_stab(test_s_blk, stab);
	
	
	while ((res = read(fd_src, word, 2*sizeof(int))) > 0) {			/* Read 64 bits from source file. */
		if ( res == 8) {										/* If we read 64 bits. */
			crypto_transformation();
			res = write(fd_dest, word, 2*sizeof(int));
			if( res != 8) {
				fprintf(stderr, "WRITE error. Try again;)\n");
				exit(1);
			}
			sourcefilesize -= 8;
		}
		else if ( sourcefilesize < 8) {							/* If we read less than 64 bits bcs source file end. */
			memset(((char*)word) + res, ' ', 8 - res);
			crypto_transformation();
			res = write(fd_dest, word, 2*sizeof(int));
			if( res != 8) {
				fprintf(stderr, "WRITE error. Try again;)\n");
				exit(1);
			}
		} else {
			fprintf(stderr, "SAD, SO SAD ERROR =(. Try again;)\n");
			exit(1);
		}
	}
	
	close(fd_key);
	close(fd_dest);
	close(fd_src);
	return 0;
}

void crypto_transformation()											/* Crypto algoritm(see GOST 28147-89). */
{
	int temp, i;
	for (i = 0; i < 24; i++) {
		temp = word[0] + GOST_key[ i % 8];
		substitution(stab, &temp);
		circl_shl(&temp);

		temp =  temp ^ word[1];
		word[1] = word[0];
		word[0] = temp;
	}
	for (i = 24; i < 31; i++) {
		temp = word[0] + GOST_key[ 31 - i];
		substitution(stab, &temp);
		circl_shl(&temp);
		temp =  temp ^ word[1];
		word[1] = word[0];
		word[0] = temp;
	}
	temp = word[0] + GOST_key[0];
	substitution(stab, &temp);
	circl_shl(&temp);
	temp =  temp ^ word[1];
	word[1] = temp;	
}

void create_stab(char src[8][16], int dest[4][256])					/* Сreate 1byte tab from 4bits tabs. */
{
	int i, j;

	for (i = 0; i < 4; i++)
		for (j = 0; j < 256; j++) 
			dest[i][j] = src[2 * i][j % 16] + 16 * src[2 * i +1][j / 16];
}

void circl_shl(int *val)
{
	unsigned int temp;
	temp = (*val);
	temp = temp>> (32 - SHIFT);
	*val = (*val) << SHIFT;
	*val = (*val) | temp;
}

void substitution(int s_tab[4][256], int *val)						//заменяет значение на сбоксы
{
	unsigned char sbox[4];
	int i, shl;
	unsigned int temp;
	for (i = 0; i < 4; i++) {
		shl = 32 - 8 * (i + 1);
		temp = (*val) << shl;
		sbox[i] = s_tab[i][temp >> 24];
	}
	*val = 0;
	for (i = 0; i< 4; i++){
		*val =*val | ( sbox[i] << 8*i);
	}
}
